from typing import List, Optional
from datetime import datetime

from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from facebook_business.adobjects.adsinsights import AdsInsights

from gcp_airflow_foundations.enums.facebook import (
    Level,
    DatePreset,
    AccountLookupScope,
    ApiObject,
)

valid_fields = {
    "account_name": AdsInsights.Field.account_name,
    "account_id": AdsInsights.Field.account_id,
    "attribution_setting": AdsInsights.Field.attribution_setting,
    "account_currency": AdsInsights.Field.account_currency,
    "campaign_name": AdsInsights.Field.campaign_name,
    "campaign_id": AdsInsights.Field.campaign_id,
    "adset_name": AdsInsights.Field.adset_name,
    "adset_id": AdsInsights.Field.adset_id,
    "ad_name": AdsInsights.Field.ad_name,
    "ad_id": AdsInsights.Field.ad_id,
    "impressions": AdsInsights.Field.impressions,
    "spend": AdsInsights.Field.spend,
    "reach": AdsInsights.Field.reach,
    "clicks": AdsInsights.Field.clicks,
    "cpc": AdsInsights.Field.cpc,
    "ctr": AdsInsights.Field.ctr,
    "cpm": AdsInsights.Field.cpm,
    "unique_clicks": AdsInsights.Field.unique_clicks,
    "inline_link_clicks": AdsInsights.Field.inline_link_clicks,
    "unique_inline_link_click_ctr": AdsInsights.Field.unique_inline_link_click_ctr,
    "inline_link_click_ctr": AdsInsights.Field.inline_link_click_ctr,
    "unique_inline_link_clicks": AdsInsights.Field.unique_inline_link_clicks,
    "cost_per_unique_inline_link_click": AdsInsights.Field.cost_per_unique_inline_link_click,
    "cost_per_unique_outbound_click": AdsInsights.Field.cost_per_unique_outbound_click,
    "cost_per_unique_click": AdsInsights.Field.cost_per_unique_click,
    "cost_per_thruplay": AdsInsights.Field.cost_per_thruplay,
    "video_30_sec_watched_actions": AdsInsights.Field.video_30_sec_watched_actions,
    "video_p25_watched_actions": AdsInsights.Field.video_p25_watched_actions,
    "video_p50_watched_actions": AdsInsights.Field.video_p50_watched_actions,
    "video_p75_watched_actions": AdsInsights.Field.video_p75_watched_actions,
    "video_p100_watched_actions": AdsInsights.Field.video_p100_watched_actions,
    "video_play_actions": AdsInsights.Field.video_play_actions,
    "conversion_values": AdsInsights.Field.conversion_values,
    "conversions": AdsInsights.Field.conversions,
    "cost_per_conversion": AdsInsights.Field.cost_per_conversion,
    "actions": AdsInsights.Field.actions,
    "action_values": AdsInsights.Field.action_values,
    "cost_per_action_type": AdsInsights.Field.cost_per_action_type,
}


@dataclass
class FacebookConfig:
    """
    Attributes:
        fields: A list of dimensions and metrics for the Facebook Graph API. For more information see: https://developers.facebook.com/docs/marketing-api/insights/parameters/v12.0
        level: Represents the level of result {ad, adset, campaign, account}
        account_lookup_scope: Whether to query all accounts managed by the user or only the active ones
        account_bq_table: A BigQuery table with the account_id's
        time_increment: The time dimension of the results
        time_range: The time range used to query the Graph API
        use_account_attribution_setting: When this parameter is set to true, your ads results will be shown using the attribution settings defined for the ad account.
        use_unified_attribution_setting: When this parameter is set to true, your ads results will be shown using unified attribution settings defined at ad set level and parameter
    """

    fields: Optional[List[str]]
    level: Optional[Level]
    account_lookup_scope: AccountLookupScope
    accounts_bq_table: Optional[str]
    time_increment: Optional[str]
    time_range: Optional[dict]
    use_account_attribution_setting: Optional[bool] = False
    use_unified_attribution_setting: Optional[bool] = False

    @validator("fields")
    def valid_fields(cls, v):
        if v is not None:
            for field in v:
                assert (
                    field in valid_fields
                ), f"`{field}` is not a valid field for the Facebook API"
            return [valid_fields[field] for field in v]
        else:
            return []



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional


@dataclass
class OdsTableMetadataConfig:
    """
    Attributes:
        hash_column_name : Hash column name
        primary_key_hash_column_name: Primary key hash column name
        ingestion_time_column_name : Insertion time column name
        update_time_column_name : Update time column name
    """

    hash_column_name: str = "af_metadata_row_hash"
    primary_key_hash_column_name: str = "af_metadata_primary_key_hash"
    ingestion_time_column_name: str = "af_metadata_inserted_at"
    update_time_column_name: str = "af_metadata_updated_at"



from dataclasses import fields
from urllib.parse import urlparse
import logging

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import (
    GCSToBigQueryOperator,
)

from gcp_airflow_foundations.base_class.salesforce_ingestion_config import (
    SalesforceIngestionConfig,
)

from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import (
    SalesforceToGcsQueryOperator,
)
from gcp_airflow_foundations.base_class.data_source_table_config import (
    DataSourceTablesConfig,
)
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class SalesforcetoBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load a CSV file from GCS to a BigQuery Table.
    """

    source_type = "SALESFORCE"

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type

    def get_bq_ingestion_task(self, dag, table_config):
        data_source = self.config.source

        # GCP Parameters
        gcp_project = data_source.gcp_project
        gcs_bucket = data_source.extra_options["gcs_bucket"]
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset

        # BigQuery parameters
        destination_table = f"{gcp_project}:{landing_dataset}.{table_config.landing_zone_table_name_override}"

        # Salesforce parameters
        object_name = table_config.extra_options.get("sf_config")["api_table_name"]
        gcs_object = f"{object_name}.csv"
        ingest_all_columns = table_config.extra_options.get("sf_config")[
            "ingest_all_columns"
        ]
        fields_to_omit = table_config.extra_options.get("sf_config")["fields_to_omit"]
        field_names = table_config.extra_options.get("sf_config")["field_names"]

        taskgroup = TaskGroup(group_id="salesforce_taskgroup")

        # 1 Load Salesforce to CSV in GCS
        gcs_upload_task = SalesforceToGcsQueryOperator(
            task_id="upload_sf_to_gcs",
            salesforce_object=object_name,
            ingest_all_fields=ingest_all_columns,
            fields_to_omit=fields_to_omit,
            fields_to_include=field_names,
            include_deleted=False,
            bucket_name=gcs_bucket,
            object_name=gcs_object,
            salesforce_conn_id="salesforce_default",
            export_format="csv",
            coerce_to_timestamp=False,
            record_time_added=True,
            gcp_conn_id="google_cloud_default",
            task_group=taskgroup,
        )

        # 2 Load CSV to BQ Landing Zone with auto-detect
        load_to_bq_landing = PythonOperator(
            task_id="import_gcs_to_bq_landing",
            op_kwargs={
                "gcs_bucket": gcs_bucket,
                "gcs_object": gcs_object,
                "destination_table": destination_table,
            },
            python_callable=self.gcs_to_bq,
            task_group=taskgroup,
        )

        gcs_upload_task >> load_to_bq_landing

        return taskgroup

    def gcs_to_bq(self, gcs_bucket, gcs_object, destination_table, **kwargs):
        ds = kwargs["ds"]

        load_to_bq = GCSToBigQueryOperator(
            task_id="import_csv_to_bq_landing",
            bucket=gcs_bucket,
            source_objects=[gcs_object],
            destination_project_dataset_table=destination_table + f"_{ds}",
            write_disposition="WRITE_TRUNCATE",
            create_disposition="CREATE_IF_NEEDED",
            skip_leading_rows=1,
        )
        load_to_bq.execute(context=kwargs)

    def validate_extra_options(self):
        tables = self.config.tables

        for table_config in tables:
            # either ingest all columns or one of field_names or fields_to_omit is non-empty
            sf_config = table_config.extra_options.get("sf_config")
            logging.info(sf_config)
            assert sf_config["ingest_all_columns"] is True or (
                sf_config["field_names"] or sf_config["fields_to_omit"]
            ), f"""Assertion error: ingest_all_columns is False and both field_names and fields_to_omit are empty for the object {table_config.table_name}"""



from gcp_airflow_foundations.enums.hds_table_type import HdsTableType


def parse_hds_schema(hds_metadata, hds_table_type, schema_fields) -> list:

    HDS_EXTRA_FIELDS_SCD2 = [
        {"name": hds_metadata.eff_start_time_column_name, "type": "TIMESTAMP"},
        {"name": hds_metadata.eff_end_time_column_name, "type": "TIMESTAMP"},
        {"name": hds_metadata.hash_column_name, "type": "STRING"},
    ]

    HDS_EXTRA_FIELDS_SNAPSHOT = [
        {"name": hds_metadata.eff_start_time_column_name, "type": "TIMESTAMP"},
        {"name": hds_metadata.partition_time_column_name, "type": "TIMESTAMP"},
        {"name": hds_metadata.hash_column_name, "type": "STRING"},
    ]

    hds_schema_fields = []

    hds_schema_fields.extend(schema_fields)

    if hds_table_type == HdsTableType.SCD2:
        hds_schema_fields.extend(HDS_EXTRA_FIELDS_SCD2)

    elif hds_table_type == HdsTableType.SNAPSHOT:
        hds_schema_fields.extend(HDS_EXTRA_FIELDS_SNAPSHOT)

    return hds_schema_fields



import json
import logging

from airflow.contrib.hooks.bigquery_hook import BigQueryHook
from airflow.contrib.hooks.gcs_hook import GoogleCloudStorageHook

from urllib.parse import urlparse


def read_schema_from_gcs(
    gcs_schema_object,
    google_cloud_storage_conn_id="google_cloud_default",
    bigquery_conn_id="google_cloud_default",
    **kwargs
) -> list:
    """
        Helper method to load table schema from a GCS URI
    """

    parsed_url = urlparse(gcs_schema_object)
    gcs_bucket = parsed_url.netloc
    gcs_object = parsed_url.path.lstrip("/")

    gcs_hook = GoogleCloudStorageHook(
        google_cloud_storage_conn_id=google_cloud_storage_conn_id, delegate_to=None
    )

    schema_fields = json.loads(
        gcs_hook.download(gcs_bucket, gcs_object).decode("utf-8")
    )

    return schema_fields



import logging
from airflow.models import BaseOperator
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.utils.decorators import apply_defaults

from gcp_airflow_foundations.operators.gcp.dlp.dlp_helpers import get_dlp_results_sql


class DlpBQInspectionResultsOperator(BaseOperator):
    """
    Reads DLP inspection results from BigQuery .

    :param project_id: GCP project ID
    :type project_id: str
    :param table_id: Target table name
    :type table_id: str
    :param dataset_id: Target dataset name
    :type dataset_id: str
    :param min_match_count: Minimum number of findings per column/likelihood level pair
    :type min_match_count: int
    """

    @apply_defaults
    def __init__(
        self,
        *,
        dataset_id,
        table_id,
        project_id,
        min_match_count=0,
        do_xcom_push=True,
        gcp_conn_id="google_cloud_default",
        impersonation_chain=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.min_match_count = min_match_count

        self.hook = BigQueryHook(
            gcp_conn_id=gcp_conn_id,
            use_legacy_sql=False,
            impersonation_chain=impersonation_chain,
        )
        conn = self.hook.get_conn()
        self.cursor = conn.cursor()

    def execute(self, context):
        sql = get_dlp_results_sql(
            self.project_id, self.dataset_id, self.table_id, self.min_match_count
        )
        results = self.hook.get_records(sql)

        logging.info(f"DLP Results are {results} ")
        if self.do_xcom_push:
            self.xcom_push(context, "results", results)
        return results



import logging
from datetime import datetime

from airflow.models.dag import DAG
from airflow.models import Variable
from airflow.utils.task_group import TaskGroup

from gcp_airflow_foundations.operators.facebook.operators.facebook_ads_to_gcs import (
    FacebookAdsReportToBqOperator,
)
from gcp_airflow_foundations.source_class.source import DagBuilder

from airflow.sensors.external_task import ExternalTaskSensor


class FacebooktoBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load Facebook Ads data to a staging BigQuery table.
    """

    source_type = "FACEBOOK"

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type

    def get_bq_ingestion_task(self, dag, table_config):
        data_source = self.config.source

        task_group = TaskGroup(group_id="ingest_facebook_data")

        facebook_options = data_source.facebook_options

        GCP_PROJECT_ID = data_source.gcp_project

        FIELDS = facebook_options.fields

        facebook_table_config = table_config.facebook_table_config
        level = facebook_options.level.value if facebook_options.level else None

        PARAMETERS = {
            "level": level,
            "time_increment": facebook_options.time_increment,
            "breakdowns": facebook_table_config.breakdowns,
            "action_breakdowns": facebook_table_config.action_breakdowns,
            "use_account_attribution_setting": facebook_options.use_account_attribution_setting,
            "use_unified_attribution_setting": facebook_options.use_unified_attribution_setting,
            "limit": "20000",
        }

        TIME_RANGE = facebook_options.time_range

        run_operator = FacebookAdsReportToBqOperator(
            task_id="fetch_facebook_data_to_bq_staging",
            api_object=facebook_table_config.api_object,
            gcp_project=GCP_PROJECT_ID,
            destination_project_dataset_table=f"{data_source.landing_zone_options.landing_zone_dataset}.{table_config.landing_zone_table_name_override}",
            accounts_bq_table=facebook_options.accounts_bq_table,
            time_range=TIME_RANGE,
            parameters=PARAMETERS,
            fields=FIELDS,
            account_lookup_scope=facebook_options.account_lookup_scope,
            gcp_conn_id="google_cloud_default",
            api_version="v12.0",
            task_group=task_group,
            dag=dag,
        )

        if data_source.external_dag_id:
            sensor_op = ExternalTaskSensor(
                task_id="wait_for_accounts_ingestion",
                external_dag_id=data_source.external_dag_id,
                mode="reschedule",
                check_existence=True,
                task_group=task_group,
                dag=dag,
            )
            sensor_op >> run_operator

        else:
            run_operator

        return task_group

    def validate_extra_options(self):
        pass



from airflow import DAG
from airflow.exceptions import AirflowException

from gcp_airflow_foundations.enums.hds_table_type import HdsTableType

from airflow.utils.task_group import TaskGroup

from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyTableOperator,
    BigQueryExecuteQueryOperator,
)

from gcp_airflow_foundations.operators.gcp.hds.hds_merge_table_operator import (
    MergeBigQueryHDS,
)
from gcp_airflow_foundations.operators.gcp.schema_migration.schema_migration_operator import (
    MigrateSchema,
)

from gcp_airflow_foundations.operators.gcp.create_table import (
    CustomBigQueryCreateEmptyTableOperator,
)


def hds_builder(
    project_id,
    table_id,
    dataset_id,
    landing_zone_dataset,
    landing_zone_table_name_override,
    column_mapping,
    column_casting,
    surrogate_keys,
    ingestion_type,
    hds_table_config,
    partition_expiration,
    location,
    dag,
    cluster_fields=None,
    time_partitioning=None,
    labels=None,
    encryption_configuration=None,
) -> TaskGroup:

    """
    Method for returning a Task Group for 1) creating an empty target HDS table (if it doesn't already exist) and for 2) for merging the staging table data into the target HDS table
    """

    taskgroup = TaskGroup(group_id="create_hds_merge_taskgroup")

    if hds_table_config.hds_table_type == HdsTableType.SNAPSHOT:
        time_partitioning = {
            "type": hds_table_config.hds_table_time_partitioning.value,
            "field": hds_table_config.hds_metadata.partition_time_column_name,
            "expirationMs": partition_expiration,
        }
    elif hds_table_config.hds_table_type == HdsTableType.SCD2:
        time_partitioning = None
    else:
        raise AirflowException(
            "Invalid HDS table type", hds_table_config.hds_table_type
        )

    # 1 Check if HDS table exists and if not create an empty table
    create_table = CustomBigQueryCreateEmptyTableOperator(
        task_id="create_hds_table",
        project_id=project_id,
        dataset_id=dataset_id,
        table_id=table_id,
        cluster_fields=cluster_fields,
        time_partitioning=time_partitioning,
        task_group=taskgroup,
        dag=dag,
    )

    # 2 Migrate schema
    migrate_schema = MigrateSchema(
        task_id="schema_migration",
        project_id=project_id,
        table_id=table_id,
        dataset_id=dataset_id,
        task_group=taskgroup,
        dag=dag,
    )

    # 3 Load staging table to HDS table
    insert = MergeBigQueryHDS(
        task_id=f"upsert_{table_id}",
        project_id=project_id,
        stg_dataset_name=landing_zone_dataset,
        data_dataset_name=dataset_id,
        stg_table_name=landing_zone_table_name_override,
        data_table_name=table_id,
        surrogate_keys=surrogate_keys,
        column_mapping=column_mapping,
        column_casting=column_casting,
        ingestion_type=ingestion_type,
        hds_table_config=hds_table_config,
        location=location,
        task_group=taskgroup,
        dag=dag,
    )

    create_table >> migrate_schema >> insert

    return taskgroup



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Optional

from paramiko.sftp import SFTP_NO_SUCH_FILE

from airflow.providers.sftp.hooks.sftp import SFTPHook
from airflow.sensors.base import BaseSensorOperator


class SFTPFilesExistenceSensor(BaseSensorOperator):
    """
    Waits for multiple file or directories to be present on SFTP.
    :param path: lit of remote files or directory paths
    :type path: str
    :param sftp_conn_id: The connection to run the sensor against
    :type sftp_conn_id: str
    """

    template_fields = ("paths",)

    def __init__(
        self, *, paths: list, sftp_conn_id: str = "sftp_default", **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.paths = paths
        self.hook: Optional[SFTPHook] = None
        self.sftp_conn_id = sftp_conn_id

    def poke(self, context: dict) -> bool:
        self.hook = SFTPHook(self.sftp_conn_id)
        for path in self.paths:
            self.log.info("Poking for %s", path)
            try:
                mod_time = self.hook.get_mod_time(path)
                self.log.info(
                    "Found File %s last modified: %s", str(path), str(mod_time)
                )
            except OSError as e:
                if e.errno != SFTP_NO_SUCH_FILE:
                    raise e
                return False
            self.hook.close_conn()
        return True



from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass
from datetime import timedelta
from datetime import datetime
from typing import List
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.enums.time_partitioning import TimePartitioning

import logging


partitioning_options = {"HOUR": "@hourly", "DAY": "@daily", "MONTH": "@monthly"}


@dataclass
class DataSourceTablesConfig:
    """
    Main configuration class used to map the fields from the configuration file.

    Attributes:
        source: The data source configuration
        tables: The table configuration
    """

    source: SourceConfig
    tables: List[SourceTableConfig]

    @root_validator(pre=True)
    def valid_partitioning(cls, values):
        ingest_schedule = values["source"].ingest_schedule

        for table in values["tables"]:
            if (
                (table.hds_config is not None)
                and (table.hds_config.hds_table_time_partitioning is not None)
                and (table.hds_config.hds_table_type == HdsTableType.SNAPSHOT)
            ):
                partitioning_time = table.hds_config.hds_table_time_partitioning.value
                assert (
                    partitioning_options[partitioning_time] == ingest_schedule
                ), f"Invalid partitioning time selection for table `{table.table_name}` - partitioning time `{partitioning_time}` must match ingestion schedule `{ingest_schedule}`"

        return values

    def dagrun_timeout(self):
        return timedelta(minutes=self.source.acceptable_delay_minutes)

    def table_start_date(self, table_config: SourceTableConfig):
        if table_config.start_date:
            start_date = table_config.start_date
        else:
            start_date = self.source.start_date

        return datetime.strptime(start_date, "%Y-%m-%d")

    def source_start_date(self):
        return datetime.strptime(self.source.start_date, "%Y-%m-%d")



from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from typing import List, Optional
import regex as re

from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType

supported_fields = ["table_name", "date"]


@dataclass
class SchemaOptionsConfig:
    schema_source_type: SchemaSourceType = SchemaSourceType.AUTO
    schema_object_template: Optional[
        str
    ] = None  # Supported tamplate fields: {table_name}, {date}

    @root_validator(pre=True)
    def valid_config(cls, values):
        if values["schema_source_type"] == SchemaSourceType.GCS:
            assert (
                values["schema_object_template"] is not None
            ), "To read schema from GCS a file template must be provided"
        return values

    @validator("schema_object_template")
    def valid_schema_object_template(cls, v):
        if v is not None:
            fields = re.findall(r"\{(.*?)\}", v)
            assert all(
                [i in supported_fields for i in fields]
            ), f"The GCS schema file template contains invalid fields: {[i for i in fields if i not in supported_fields]}"
        return v



from airflow.hooks.base import BaseHook
from typing import Optional
from airflow.exceptions import AirflowException

import requests
import time
import json
import logging


class DataformHook(BaseHook):
    """
    Airflow Hook to connect to Dataform's API and run Dataform jobs.
    This hook will POST to Dataform's ApiService_RunCreate to initiate a new dataform run.
    If no tags are provided, Dataform will run all.
    Once a run is created, it will use ApiService_RunGet to return information about the run every 10 seconds until the status is no longer RUNNING.

    Attributes:
    :param environment: name of the environment, e.g. production, development...
    :type environment: str
    :param schedule: the name of the schedule that will be triggered by airflow
    :type schedule: str
    :param tags: (optional) A list of tags with which the action must have in order to be run. If not set, then all actions will run.
    :type tags: list[str]
    :param wait_until_finished: (optional) ait until job finishes running, either return success or error
    :type wait_until_finished: bool
    :param sleep_time: (optional) if status returns "RUNNING", wait for configured time before retry. Default is 60 seconds
    :type sleep_time: int

    Instructions to prepare Dataform for the API call:
    1. create schedule in Dataform for REST API call
        a) environments.json >> under the appropriate environment (e.g. prod, dev..) click on "create new schedule"
        b) enter schedule name (e.g. composer)
        c) disable "Enable this schedule"
    2. get Dataform project ID
        easiest way is to look at the url
        E.g. https://app.dataform.co/#/<project_id>/b/<branch>/file/<file-name>
    3. generate API token:
        project settings >> API keys >> GENERATE NEW API KEY



    - Dataform documentation on using REST API: https://docs.dataform.co/dataform-web/api
    - Helpful Medium tutorial: https://medium.com/google-cloud/cloud-composer-apache-airflow-dataform-bigquery-de6e3eaabeb3

    ** Note: Schedules must be on the master branch. In Dataform, you'll have to create a branch first and then merge changes into master.
    """

    def __init__(self, dataform_conn_id="datafrom_default") -> None:
        self.conn = self.get_connection(dataform_conn_id)
        self.project_id = self.conn.login
        self.api_key = self.conn.password

    def run_job(
        self,
        environment: str,
        schedule: str,
        tags: Optional[str] = [],
        wait_until_finished: bool = True,
        sleep_time: int = 60,
    ) -> str:
        base_url = f"https://api.dataform.co/v1/project/{self.project_id}/run"
        headers = {"Authorization": f"Bearer {self.api_key}"}
        run_create_request = {
            "environmentName": environment,
            "scheduleName": schedule,
            "runConfig": {"tags": tags},
        }

        # ApiService_RunCreate: Initiates new dataform runs within the project, and returns the ID of any created runs.
        response = requests.post(
            base_url, data=json.dumps(run_create_request), headers=headers
        )

        try:
            # ApiService_RunGet: Returns information about a specific run
            run_url = base_url + "/" + response.json()["id"]
        except Exception:
            raise AirflowException(f"Dataform ApiService Error: {response.json()}")

        while wait_until_finished:
            response = requests.get(run_url, headers=headers)

            if response.json()["status"] == "RUNNING":
                response = requests.get(run_url, headers=headers)
                logging.info(f"Dataform run in progress: {response.json()}")
                time.sleep(sleep_time)
            elif response.json()["status"] == "SUCCESSFUL":
                return f"Dataform run completed: SUCCESSFUL see run logs at {response.json()['runLogUrl']}"
            else:
                raise AirflowException(
                    f"Dataform run {response.json()['status']} for {response.json()['id']}: see run logs at {response.json()['runLogUrl']}"
                )



from airflow.models import BaseOperator
from airflow.models.dag import DAG
from airflow.operators.dummy import DummyOperator
from airflow.utils.trigger_rule import TriggerRule

from gcp_airflow_foundations.base_class.dlp_table_config import DlpTableConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.operators.gcp.delete_staging_table import (
    BigQueryDeleteStagingTableOperator,
)
from gcp_airflow_foundations.operators.gcp.dlp.dlp_to_datacatalog_taskgroup import (
    schedule_dlp_to_datacatalog_taskgroup_multiple_tables,
)
from gcp_airflow_foundations.operators.gcp.hds.load_hds_taskgroup import hds_builder
from gcp_airflow_foundations.operators.gcp.ods.load_ods_taskgroup import ods_builder
from gcp_airflow_foundations.operators.gcp.schema_parsing.schema_parsing_operator import (
    ParseSchema,
)
from gcp_airflow_foundations.source_class.schema_source_config import SchemaSourceConfig


def load_builder(
    data_source: SourceConfig,
    table_config: SourceTableConfig,
    schema_config: SchemaSourceConfig,
    preceding_task: BaseOperator,
    dag: DAG,
):

    """
    Method for building all needed Task Groups based on the HdsTableConfig and OdsTableConfig options declared in the config files
    Takes data from landing-zone/stating-area and ingests it into the Data Lake.
    """

    project_id = data_source.gcp_project
    dataset_id = data_source.dataset_data_name
    dataset_hds_id = data_source.dataset_hds_override
    landing_zone_dataset = data_source.landing_zone_options.landing_zone_dataset
    landing_zone_table_name_override = table_config.landing_zone_table_name_override
    surrogate_keys = table_config.surrogate_keys
    column_mapping = table_config.column_mapping
    column_casting = table_config.column_casting
    ingestion_type = table_config.ingestion_type
    partition_expiration = data_source.partition_expiration
    ods_table_config = table_config.ods_config
    hds_table_config = table_config.hds_config
    location = data_source.location
    cluster_fields = table_config.cluster_fields
    dlp_table_config = DlpTableConfig()
    dlp_table_config.set_source_config(data_source.dlp_config)
    ods_suffix = data_source.ods_suffix
    hds_suffix = data_source.hds_suffix

    ods_table_config.table_id = f"{landing_zone_table_name_override}{ods_suffix}"

    parse_schema = ParseSchema(
        task_id="schema_parsing",
        schema_config=schema_config,
        column_mapping=column_mapping,
        column_casting=column_casting,
        data_source=data_source,
        table_config=table_config,
        dag=dag,
    )

    ods_task_group = ods_builder(
        project_id=project_id,
        table_id=ods_table_config.table_id,
        dataset_id=dataset_id,
        landing_zone_dataset=landing_zone_dataset,
        landing_zone_table_name_override=landing_zone_table_name_override,
        surrogate_keys=surrogate_keys,
        column_mapping=column_mapping,
        column_casting=column_casting,
        ingestion_type=ingestion_type,
        cluster_fields=cluster_fields,
        partition_expiration=partition_expiration,
        ods_table_config=ods_table_config,
        location=location,
        dag=dag,
    )

    hds_task_group = None
    if hds_table_config:
        hds_table_config.table_id = f"{landing_zone_table_name_override}{hds_suffix}"

        hds_task_group = hds_builder(
            project_id=project_id,
            table_id=hds_table_config.table_id,
            dataset_id=dataset_hds_id,
            landing_zone_dataset=dataset_id,
            landing_zone_table_name_override=ods_table_config.table_id,
            surrogate_keys=surrogate_keys,
            column_mapping=column_mapping,
            column_casting=column_casting,
            ingestion_type=ingestion_type,
            cluster_fields=cluster_fields,
            partition_expiration=partition_expiration,
            hds_table_config=hds_table_config,
            location=location,
            dag=dag,
        )

    delete_staging_table = BigQueryDeleteStagingTableOperator(
        task_id="delete_staging_table",
        project_id=project_id,
        dataset_id=landing_zone_dataset,
        table_id=landing_zone_table_name_override,
        dag=dag,
    )

    if hds_task_group:
        preceding_task >> parse_schema >> ods_task_group >> hds_task_group >> delete_staging_table
    else:
        preceding_task >> parse_schema >> ods_task_group >> delete_staging_table

    done = DummyOperator(task_id="done", trigger_rule=TriggerRule.ALL_DONE)

    if dlp_table_config.get_is_on():
        dlp_tasks_configs = [
            {
                "datastore": "ods",
                "project_id": project_id,
                "table_id": ods_table_config.table_id,
                "dataset_id": dataset_id,
            }
        ]
        if hds_table_config:
            dlp_tasks_configs.append(
                {
                    "datastore": "hds",
                    "project_id": project_id,
                    "table_id": hds_table_config.table_id,
                    "dataset_id": dataset_hds_id,
                }
            )

        ods_dlp_task_groups = schedule_dlp_to_datacatalog_taskgroup_multiple_tables(
            table_configs=dlp_tasks_configs,
            table_dlp_config=dlp_table_config,
            next_task=done,
            dag=dag,
        )

        delete_staging_table >> ods_dlp_task_groups >> done
    else:
        delete_staging_table >> done



from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
from abc import ABC, abstractmethod, abstractproperty
import logging
import json
import pandas as pd
from dacite import from_dict
from dataclasses import dataclass
from datetime import datetime

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import (
    GCSToBigQueryOperator,
)
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.exceptions import AirflowException

from gcp_airflow_foundations.base_class.gcs_table_config import GCSTableConfig
from gcp_airflow_foundations.base_class.gcs_source_config import GCSSourceConfig
from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import (
    SalesforceToGcsQueryOperator,
)
from gcp_airflow_foundations.operators.api.sensors.gcs_sensor import (
    GCSObjectListExistenceSensor,
)
from gcp_airflow_foundations.operators.api.sensors.gcs_prefix_sensor import (
    GCSObjectPrefixListExistenceSensor,
)
from gcp_airflow_foundations.source_class.ftp_source import FTPtoBQDagBuilder


class GCSFiletoBQDagBuilder(FTPtoBQDagBuilder):
    """
    Builds DAGs to load files from GCS to a BigQuery Table.

    For GCS->BQ ingestion, either a metadata file is required or the field templated_file_name must be provided.
    If a metadata file is provided, itt can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest
    """

    source_type = "GCS"

    def metadata_file_sensor(self, table_config, taskgroup):
        """
        Implements a sensor for the metadata file specified in the table config.
        """
        logging.info(
            "table_config.extra_options {} ".format(table_config.extra_options)
        )
        if "metadata_file" in table_config.extra_options.get("gcs_table_config"):
            metadata_file_name = table_config.extra_options.get("gcs_table_config")[
                "metadata_file"
            ]
            bucket = self.config.source.extra_options["gcs_bucket"]

            return GCSObjectExistenceSensor(
                task_id="wait_for_metadata_file",
                bucket=bucket,
                object=metadata_file_name,
                task_group=taskgroup,
            )
        else:
            return DummyOperator(
                task_id="dummy_metadata_file_ingestion", task_group=taskgroup
            )

    def file_ingestion_task(self, table_config, taskgroup):
        """
        No ingestion is needed - data is already in GCS, so return a dummy operator.
        """
        return DummyOperator(
            task_id="dummy_file_ingestion_operator", task_group=taskgroup
        )

    def schema_file_sensor(self, table_config, taskgroup):
        """
        Implements an Airflow sensor to wait for an (optional) schema file in GCS
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        schema_file_name = None
        if "schema_file" in table_config.extra_options.get("gcs_table_config"):
            schema_file_name = table_config.extra_options.get("gcs_table_config")[
                "schema_file"
            ]
        logging.info(schema_file_name)

        if schema_file_name:
            return GCSObjectExistenceSensor(
                task_id="wait_for_schema_file",
                bucket=bucket,
                object=schema_file_name,
                task_group=taskgroup,
            )
        else:
            return DummyOperator(
                task_id="dummy_schema_file_ingestion_operator", task_group=taskgroup
            )

    def get_file_list_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="get_file_list",
            op_kwargs={"table_config": table_config},
            python_callable=self.get_list_of_files,
            task_group=taskgroup,
        )

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        files_to_wait_for = "{{ ti.xcom_pull(key='file_list', task_ids='ftp_taskgroup.get_file_list') }}"

        logging.info(self.config.source.extra_options)
        if self.config.source.extra_options["gcs_source_config"][
            "file_prefix_filtering"
        ]:
            return GCSObjectPrefixListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                prefixes=files_to_wait_for,
                task_group=taskgroup,
            )
        else:
            return GCSObjectListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                objects=files_to_wait_for,
                task_group=taskgroup,
            )

    def load_to_landing_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="load_gcs_to_landing_zone",
            op_kwargs={"table_config": table_config},
            python_callable=self.load_to_landing_py_op_task,
            task_group=taskgroup,
        )

    def load_to_landing_py_op_task(self, table_config, **kwargs):
        data_source = self.config.source
        bucket = self.config.source.extra_options["gcs_bucket"]

        ti = kwargs["ti"]
        ds = kwargs["ds"]

        files_to_load = ti.xcom_pull(
            key="file_list", task_ids="ftp_taskgroup.get_file_list"
        )
        logging.info(files_to_load)

        if self.config.source.extra_options["gcs_source_config"][
            "file_prefix_filtering"
        ]:
            logging.info("HELLO")
            for i in range(len(files_to_load)):
                matching_gcs_files = self.gcs_hook.list(
                    bucket_name=bucket, prefix=files_to_load[i]
                )
                if len(matching_gcs_files) > 1:
                    raise AirflowException(
                        f"There is more than one matching file with the prefix {files_to_load[i]} in the bucket {bucket}"
                    )
                files_to_load[i] = matching_gcs_files[0]

        logging.info(files_to_load)

        # Parameters
        bucket = self.config.source.extra_options["gcs_bucket"]
        source_format = self.config.source.extra_options["gcs_source_config"][
            "source_format"
        ]
        field_delimeter = self.config.source.extra_options["gcs_source_config"][
            "delimeter"
        ]
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset
        destination_table = (
            f"{gcp_project}:{landing_dataset}.{table_config.landing_zone_table_name_override}"
            + f"_{ds}"
        )

        schema_file_name = None
        if "schema_file" in table_config.extra_options.get("gcs_table_config"):
            schema_file_name = table_config.extra_options.get("gcs_table_config")[
                "schema_file"
            ]
        logging.info(field_delimeter)

        # Get files to load from metadata file
        if schema_file_name:
            schema_file = self.gcs_hook.download(
                bucket_name=bucket, object_name=schema_file_name
            )

            # Only supports json schema file format - add additional support if required
            schema_fields = json.loads(schema_file)
            logging.info(schema_fields)
            gcs_to_bq = GCSToBigQueryOperator(
                task_id="import_files_to_bq_landing",
                bucket=bucket,
                source_objects=files_to_load,
                source_format=source_format,
                schema_fields=schema_fields,
                field_delimeter=field_delimeter,
                destination_project_dataset_table=destination_table,
                write_disposition="WRITE_TRUNCATE",
                create_disposition="CREATE_IF_NEEDED",
                skip_leading_rows=1,
            )
        else:
            gcs_to_bq = GCSToBigQueryOperator(
                task_id="import_files_to_bq_landing",
                bucket=bucket,
                source_objects=files_to_load,
                source_format=source_format,
                field_delimiter=field_delimeter,
                destination_project_dataset_table=destination_table,
                write_disposition="WRITE_TRUNCATE",
                create_disposition="CREATE_IF_NEEDED",
                skip_leading_rows=1,
            )
        gcs_to_bq.execute(context=kwargs)

    def get_list_of_files(self, table_config, **kwargs):
        # XCom push the list of files
        if "metadata_file" in table_config.extra_options.get("gcs_table_config"):
            file_list = []
            with open("metadata.csv", newline="") as f:
                for line in f:
                    file_list.append(line.strip())
        else:
            templated_file_name = self.config.source.extra_options["gcs_source_config"][
                "templated_file_name"
            ]
            templated_file_name = templated_file_name.replace(
                "{{ TABLE_NAME }}", table_config.table_name
            )
            file_list = [templated_file_name]

        # support replacing files with current dates
        ds = kwargs["ds"]
        ds = datetime.strptime(ds, "%Y-%m-%d").strftime(
            self.config.source.extra_options["gcs_source_config"]["date_format"]
        )
        file_list[:] = [
            file.replace("{{ ds }}", ds) if "{{ ds }}" in file else file
            for file in file_list
        ]

        # add dir prefix to files
        x = "directory_prefix" in self.config.source.extra_options["gcs_source_config"]
        logging.info(x)
        if "directory_prefix" in self.config.source.extra_options["gcs_source_config"]:
            dir_prefix = self.config.source.extra_options["gcs_source_config"][
                "directory_prefix"
            ]
            file_list[:] = [dir_prefix + file for file in file_list]
        logging.info(file_list)

        kwargs["ti"].xcom_push(key="file_list", value=file_list)

    def validate_extra_options(self):
        pass
        # gcs_source_cfg = from_dict(data_class=GCSSourceConfig, data=self.config.source.extra_options["gcs_source_config"])
        # tables = self.config.tables
        # for table in tables:
        #    gcs_table_cfg = from_dict(data_class=GCSTableConfig, data=table.extra_options.get("gcs_table_config"))



import requests
from airflow.hooks.base import BaseHook
import logging


class MailgunHook(BaseHook):
    """
    Mailgun Hook for validating email addresses.
    API payload format example:
    {
        'address': 'zzzzzzzzz@gmail.com',
        'is_disposable_address': False,
        'is_role_address': False,
        'reason': [],
        'result': 'deliverable',
        'risk': 'low'
    }
    """

    def __init__(self, mailgun_conn_id: str) -> dict:
        self.conn = self.get_connection(mailgun_conn_id)
        self.session = requests.Session()
        self.session.auth = ("api", self.conn.login)

    def validate_email(self, email):
        res = self.session.get(
            "https://api.mailgun.net/v4/address/validate", params={"address": email}
        )
        return res.json()

    def validate_email_bulk(self, file, job_id):
        res = self.session.post(
            f"https://api.mailgun.net/v4/address/validate/bulk/{job_id}",
            files={"file": open(file, "rb")},
        )
        return res.json()

    def check_job_status(self, job_id):
        res = self.session.get(
            f"https://api.mailgun.net/v4/address/validate/bulk/{job_id}"
        )
        logging.info(f"job status: {res.json()}")
        return res

    def download_csv(self, job_id):
        res_job = self.session.get(
            f"https://api.mailgun.net/v4/address/validate/bulk/{job_id}"
        )
        csv_link = res_job.json()["download_url"]["csv"]
        return csv_link



import datetime
from typing import List, Optional
from croniter import croniter
from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from gcp_airflow_foundations.enums.source_type import SourceType
from gcp_airflow_foundations.base_class.landing_zone_config import LandingZoneConfig
from gcp_airflow_foundations.base_class.schema_options_config import SchemaOptionsConfig
from gcp_airflow_foundations.base_class.facebook_config import FacebookConfig
from gcp_airflow_foundations.base_class.dlp_source_config import DlpSourceConfig

partition_limit = 4000
ms_day = 86400000
expiration_options = {
    "@hourly": partition_limit / 24,
    "@daily": partition_limit,
    "@monthly": partition_limit * 30,
}


@dataclass
class SourceConfig:
    """
    Source configuration data class.

    Attributes:
        name : Name of source
        source_type : Source type selection. See SourceType class
        ingest_schedule : Ingestion schedule. Currently only supporting @hourly, @daily, @weekly, and @monthly
        gcp_project : Google Cloud Platform project ID
        dataset_data_name : Target dataset name
        extra_options : Google Cloud Storage bucket and objects for source data if loading from GCS
        landing_zone_options : Staging dataset name
        acceptable_delay_minutes : Delay minutes limit
        notification_emails : Email address for notification emails
        owner : Airflow user owning the DAG
        partition_expiration: Expiration time for HDS Snapshot partitions in days.
        facebook_options: Extra options for ingesting data from Facebook Marketing API.
        catchup: Run all dag runs since start_date. https://airflow.apache.org/docs/apache-airflow/stable/dag-run.html#catchup
        dag_args: Optional dictionary of parameters to be passed as keyword arguments to the ingestion DAG.
                    Refer to :class:`airflow.models.dag.DAG` for the available parameters.
        location: BigQuery job location.
        start_date : Start date for DAG
        start_date_tz : Timezone
        ods_suffix : Suffix for ODS tables. Defaults to empty string.
        hds_suffix : Suffix for HDS tables. Defaults to empty string.
        version : The Dag version. Can be incremented if logic changes
        sla_mins : Service Level Agreement (SLA) timeout minutes. This is is an expectation for the maximum time a Task should take.
        connection  : Aiflow Google Cloud Platform connection

    """

    name: str
    source_type: str
    ingest_schedule: str
    external_dag_id: Optional[str]
    gcp_project: str
    dataset_data_name: str
    dataset_hds_override: Optional[str]
    extra_options: Optional[dict]
    landing_zone_options: LandingZoneConfig
    landing_zone_options: LandingZoneConfig
    acceptable_delay_minutes: int
    notification_emails: List[str]
    owner: str
    partition_expiration: Optional[int]
    dag_args: Optional[dict]
    location: str
    start_date: str
    schema_options: SchemaOptionsConfig = SchemaOptionsConfig()
    facebook_options: Optional[FacebookConfig] = None  # TODO: Move into extra_configs
    catchup: bool = True
    start_date_tz: str = "EST"
    ods_suffix: str = ""
    hds_suffix: str = ""
    version: int = 1
    sla_mins: int = 900
    dlp_config: DlpSourceConfig = None
    connection: str = "google_cloud_default"

    @validator("name")
    def valid_name(cls, v):
        assert v, "Source name must not be empty"
        assert "." not in v, "Source Name cannot contain the period character"
        return v

    @validator("source_type")
    def valid_source_type(cls, v):
        assert v, "Source type must not be empty"
        return v

    @validator("ingest_schedule")
    def valid_ingest_schedule(cls, v):
        assert v in ["@hourly", "@daily", "@weekly", "@monthly"] or croniter.is_valid(
            v
        ), "invalid ingest schedule: see Airflow documentation for more details"
        return v

    @validator("landing_zone_options")
    def valid_landing_zone(cls, v):
        assert v.landing_zone_dataset, "Landing zone dataset name must not be empty"
        return v

    @validator("start_date")
    def valid_start_date(cls, v):
        assert datetime.datetime.strptime(
            v, "%Y-%m-%d"
        ), "The date format for Start Date should be YYYY-MM-DD"
        return v

    @validator("schema_options")
    def valid_schema_options(cls, v):
        assert v is not None, "Schema options configuration must be provided"
        return v

    @root_validator(pre=True)
    def valid_partition_expiration(cls, values):
        if values["partition_expiration"] is not None:
            assert (
                values["partition_expiration"]
                < expiration_options[values["ingest_schedule"]]
            ), f"The partition limit should be smaller than {expiration_options[values['ingest_schedule']]} days. It is currently set to {values['partition_expiration']}"
            values["partition_expiration"] = values["partition_expiration"] * ms_day
        return values

    @root_validator(pre=True)
    def valid_hds_dataset(cls, values):
        if values["dataset_hds_override"] is None:
            values["dataset_hds_override"] = values["dataset_data_name"]
        return values

    @root_validator(pre=True)
    def valid_table_suffix(cls, values):
        if values["hds_suffix"] == values["ods_suffix"]:
            assert (
                values["dataset_hds_override"] is not None
            ), "ODS and HDS table must reside in separate datasets if they share the same suffix."
        return values



from airflow.models import Variable

def save_id_to_file(private_key_var):
    key = Variable.get(private_key_var)
    secret_path = "id_rsa"
    with open(secret_path, "w") as f:
        f.write(key)


from typing import Optional

from airflow.models import BaseOperator, BaseOperatorLink
from airflow.contrib.operators.bigquery_operator import (
    BigQueryOperator,
    BigQueryCreateEmptyTableOperator,
)

from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

from airflow.exceptions import AirflowException

import logging

from gcp_airflow_foundations.common.gcp.source_schema.gcs import read_schema_from_gcs
from gcp_airflow_foundations.common.gcp.ods.schema_utils import parse_ods_schema
from gcp_airflow_foundations.common.gcp.hds.schema_utils import parse_hds_schema


from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig
from gcp_airflow_foundations.base_class.hds_table_config import HdsTableConfig


class ParseSchema(BaseOperator):
    @apply_defaults
    def __init__(
        self,
        *,
        schema_config,
        column_mapping=None,
        column_casting=None,
        data_source=None,
        table_config=None,
        **kwargs,
    ) -> list:
        super().__init__(**kwargs)

        self.schema_config = schema_config
        self.column_mapping = column_mapping
        self.column_casting = column_casting
        self.data_source = data_source
        self.table_config = table_config

        self.ods_table_config = table_config.ods_config
        self.hds_table_config = table_config.hds_config

    def execute(self, context):
        ds = context["ds"]

        schema_source_config_class = self.schema_config
        schema_method = schema_source_config_class().schema_method()
        schema_method_arguments = schema_source_config_class().schema_method_arguments(
            self.data_source, self.table_config, ds=ds
        )

        source_schema_fields = schema_method(**schema_method_arguments)
        logging.info(f"Parsed schema using: {schema_method}")

        schema_xcom = {
            "source_table_columns": [field["name"] for field in source_schema_fields]
        }

        if self.column_casting:
            for field in source_schema_fields:
                if field["name"] in self.column_casting:
                    field["type"] = self.column_casting[field["name"]]["output_type"]

        if self.column_mapping:
            for field in source_schema_fields:
                if field["name"] in self.column_mapping:
                    field["name"] = self.column_mapping[field["name"]]

        if self.ods_table_config:
            schema_xcom[
                f"{self.data_source.dataset_data_name}.{self.ods_table_config.table_id}"
            ] = parse_ods_schema(
                schema_fields=source_schema_fields,
                ods_metadata=self.ods_table_config.ods_metadata,
            )

        if self.hds_table_config:
            schema_xcom[
                f"{self.data_source.dataset_hds_override}.{self.hds_table_config.table_id}"
            ] = parse_hds_schema(
                schema_fields=source_schema_fields,
                hds_metadata=self.hds_table_config.hds_metadata,
                hds_table_type=self.hds_table_config.hds_table_type,
            )

        return schema_xcom



from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field
from pydantic import validator, root_validator

from datetime import datetime
from typing import List, Optional

from gcp_airflow_foundations.base_class.hds_metadata_config import (
    HdsTableMetadataConfig,
)
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.enums.time_partitioning import TimePartitioning


@dataclass
class HdsTableConfig:
    """
    Attributes:
        hds_table_type : SNAPSHOT OR SCD2
        hds_table_time_partitioning: Partitioning for BigQuery table. One of DAY, HOUR, or MONTH
        hds_metadata : See HdsTableMetadataConfig class
    """

    hds_table_time_partitioning: Optional[TimePartitioning]  # DAY, HOUR, or MONTH
    hds_table_type: HdsTableType = HdsTableType.SCD2  # SNAPSHOT OR SCD2
    hds_metadata: HdsTableMetadataConfig = HdsTableMetadataConfig()

    @root_validator(pre=True)
    def valid_config(cls, values):
        if values["hds_table_type"] == HdsTableType.SNAPSHOT:
            assert (
                values["hds_table_time_partitioning"] is not None
            ), "HDS snapshot table's partition time must match the ingestion schedule"
        return values



from airflow.models.baseoperator import BaseOperator
from gcp_airflow_foundations.operators.api.hooks.dataform_hook import DataformHook
from airflow.utils.decorators import apply_defaults
from typing import Any, Optional


class DataformOperator(BaseOperator):
    """
    This operator will use the custom DataformHook to POST to Dataform's ApiService_RunCreate to initiate a new dataform run.
    If no tags are provided, Dataform will run all.
    Once a run is created, it will use ApiService_RunGet to return information about the run every 10 seconds until the status is no longer RUNNING.

    Attributes:
    :param environment: name of the environment, e.g. production, development...
    :type environment: str
    :param schedule: the name of the schedule that will be triggered by airflow
    :type schedule: str
    :param tags: (optional) A list of tags with which the action must have in order to be run. If not set, then all actions will run.
    :type tags: list[str]
    :param wait_until_finished: (optional) wait until job finishes running, either return success or error
    :type wait_until_finished: bool

    Instructions to prepare Dataform for the API call:
    1. create schedule in Dataform for REST API call
        a) environments.json >> under the appropriate environment (e.g. prod, dev..) click on "create new schedule"
        b) enter schedule name (e.g. composer)
        c) disable "Enable this schedule"
    2. get Dataform project ID
        easiest way is to look at the url
        E.g. https://app.dataform.co/#/<project_id>/b/<branch>/file/<file-name>
    3. generate API token:
        project settings >> API keys >> GENERATE NEW API KEY

    - Dataform documentation on using REST API: https://docs.dataform.co/dataform-web/api
    - Helpful Medium tutorial: https://medium.com/google-cloud/cloud-composer-apache-airflow-dataform-bigquery-de6e3eaabeb3

    ** Note: Schedules must be on the master branch. In Dataform, you'll have to create a branch first and then merge changes into master.
    """

    @apply_defaults
    def __init__(
        self,
        *,
        dataform_conn_id: str = "dataform_default",
        environment: str,
        schedule: str,
        tags: Optional[str] = [],
        wait_until_finished: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.dataform_conn_id = dataform_conn_id
        self.environment = environment
        self.schedule = schedule
        self.tags = tags
        self.wait_until_finished = wait_until_finished

    def execute(self, context) -> str:
        dataform_hook = DataformHook(dataform_conn_id=self.dataform_conn_id)
        dataform_hook.run_job(
            environment=self.environment,
            schedule=self.schedule,
            tags=self.tags,
            wait_until_finished=self.wait_until_finished,
        )
